#define MLOG_FACSARRAY /* need to define before include mlog .h files */
#include <mlogfacs.h>

#include <iod_internal.h>

enum iod_tokenize_rc {
	TOKEN_OK,
	TOKEN_BLANK_LINE,
	TOKEN_INVALID_KEY,
	TOKEN_NO_VALUE,
};

static int tokenize(char *line, char **token, char **value)
{
	char *p;

	for (p = line; isspace(*p); p++) ;
	if (*p == '\0' || *p == '#') {
		return TOKEN_BLANK_LINE;
	}

	*token = p;
	if (isdigit(*p)) {
		for (; !isspace(*p) && *p != '\0'; p++) ;
		*p = '\0';
		return TOKEN_INVALID_KEY;
	}
	for (; isalnum(*p) || *p == '_'; p++) ;
	if (!isspace(*p)) {
		for (; !isspace(*p) && *p != '\0'; p++) ;
		*p = '\0';
		return TOKEN_INVALID_KEY;
	}

	*p = '\0';
	for (p++; isspace(*p); p++) ;
	if (*p == '\0' || *p == '#') {
		return TOKEN_NO_VALUE;
	}

	*value = p; /* value will be parsed later according to its type */
	for (p++; !isspace(*p); p++);
	*p = '\0';
	return TOKEN_OK;
}

static int parse_conf_keyval(iod_conf_t *pconf, char *key, char *value)
{
	int rc = 0;
	int v;
	if (strcmp(key, "mlog_stderr") == 0) {
		v = atoi(value);
		if (v) {
			pconf->mlog_flags |= MLOG_STDERR;
		} else {
			pconf->mlog_flags &= ~MLOG_STDERR;
		}
	} else if (strcmp(key, "mlog_ucon") == 0) {
		v = atoi(value);
		if (v) {
			pconf->mlog_flags |= (MLOG_UCON_ON | MLOG_UCON_ENV);
		} else {
			pconf->mlog_flags &= ~(MLOG_UCON_ON | MLOG_UCON_ENV);
		}
	} else if (strcmp(key, "mlog_syslog") == 0) {
		v = atoi(value);
		if (v) {
			pconf->mlog_flags |= MLOG_SYSLOG;
		} else {
			pconf->mlog_flags &= ~MLOG_SYSLOG;
		}
	} else if (strcmp(key, "mlog_defmask") == 0) {
		pconf->mlog_defmask = mlog_str2pri(value);
		if (pconf->mlog_defmask < 0) {
			mlog(MLOG_DBG, "Bad mlog_defmask value: %d",
				pconf->mlog_defmask);
			rc = -EINVAL;
		}
	} else if (strcmp(key, "mlog_stderrmask") == 0) {
		pconf->mlog_stderrmask = mlog_str2pri(value);
		if (pconf->mlog_stderrmask < 0) {
			mlog(MLOG_DBG, "Bad mlog_stderrmask value");
			rc = -EINVAL;
		}
	} else if (strcmp(key, "mlog_file") == 0) {
		pconf->mlog_file = strdup(value);
	} else if (strcmp(key, "mlog_msgbuf_size") == 0) {
		pconf->mlog_msgbuf_size = atoi(value);
		/*
		 * 0 means disable it.  negative non-zero values or very
		 * small positive numbers don't make sense, so disallow.
		 */
		if (pconf->mlog_msgbuf_size < 0 ||
		    (pconf->mlog_msgbuf_size > 0 &&
		     pconf->mlog_msgbuf_size < 256)) {
			mlog(MLOG_DBG, "Unknown key: %s", key);
			rc = -EINVAL;
		}
	} else if (strcmp(key, "mlog_setmasks") == 0) {
		pconf->mlog_setmasks = strdup(value);
		if (pconf->mlog_setmasks == NULL) {
			/*
			 * XXX: strdup fails, new will too, and we don't handle
			 * exceptions... so we'll assert here.
			 */
			mlog(MLOG_DBG, "Unable to malloc mlog_setmasks");
			rc = -EINVAL;
		}
	} else {
		mlog(MLOG_DBG, "Unknown key: %s", key);
		rc = -EINVAL;
	}
	return rc;
}

static iod_conf_t *parse_conf(FILE * fp, const char *file, iod_conf_t *pconf)
{
	char input[IOD_CONFIG_LINE_MAX], *key, *value;
	int line;
	int rc = 0;
	char *errormsg;

	errormsg = (char *) calloc(1, 1024);
	if (errormsg == NULL) {
		mlog(MLOG_DBG, "Out of memory");
		return NULL;
	}
	if (pconf == NULL) {
		pconf = (iod_conf_t *)calloc(1, sizeof(iod_conf_t));
		if (pconf == NULL) {
			mlog(MLOG_DBG, "Out of memory");
			free(errormsg);
		}
	}
	mlog(MLOG_DBG, "Parsing %s", file);
	for (line = 1; fgets(input, IOD_CONFIG_LINE_MAX, fp); line++) {
		switch (tokenize(input, &key, &value)) {
		case TOKEN_BLANK_LINE:
			continue;
		case TOKEN_INVALID_KEY:
			sprintf(errormsg, "invalid key: %s", key);
			rc = -EINVAL;
			break;
		case TOKEN_NO_VALUE:
			sprintf(errormsg, "no value for key: %s", key);
			rc = -EINVAL;
			break;
		case TOKEN_OK:
			break;
		}
		if (rc == 0) {
			rc = parse_conf_keyval(pconf, key, value);
			if (rc != 0)
				break;
		} else {
			break;
		}
		mlog(MLOG_DBG, "Read (%d) %s %s", line, key, value);
	}
	if (rc != 0) {
		mlog(MLOG_WARN, "Error in the conf file %s (%d): %s",
		     file, line, errormsg);
		free(pconf);
	} else {
		mlog(MLOG_DBG, "Successfully parsed conf file %s", file);
		return pconf;
	}

	free(errormsg);
	return NULL;
}

/**
 * setup_mlog_facnamemask: setup the mlog facility names and inital
 * mask.    helper function for setup_mlog() and get_plfs_conf(), the
 * latter for the early mlog init before the plfsrc is read.
 *
 * @param masks masks in mlog_setmasks() format, or NULL
 */
void
setup_mlog_facnamemask(char *masks)
{
    int lcv;
    /* name facilities */
    for (lcv = 0; mlog_facsarray[lcv] != NULL ; lcv++) {
        /* can't fail, as we preallocated in mlog_open() */
        if (lcv == 0) {
            continue;    /* don't mess with the default facility */
        }
        (void) mlog_namefacility(lcv, (char *)mlog_facsarray[lcv],
                                 (char *)mlog_lfacsarray[lcv]);
    }
    /* finally handle any mlog_setmasks() calls */
    if (masks != NULL) {
        mlog_setmasks(masks, -1);
    }
}

static void
setup_mlog(iod_conf_t *pconf)
{
	char *level, *ev, *start;
	char tmpbuf[64];

	/* end of part 1 of simplified high-level env var config */
	/* shutdown early mlog config so we can replace with the real one ... */
	mlog_close();
	/* now we are ready to mlog_open ... */
	if (mlog_open((char *)"iod",
		/* don't count the null at end of mlog_facsarray */
		sizeof(mlog_facsarray)/sizeof(mlog_facsarray[0]) - 1,
		pconf->mlog_defmask, pconf->mlog_stderrmask,
		pconf->mlog_file, pconf->mlog_msgbuf_size,
		pconf->mlog_flags, pconf->mlog_syslogfac) < 0) {
		fprintf(stderr, "mlog_open: failed.  Check mlog params.\n");
		exit(1);
	}

	setup_mlog_facnamemask(pconf->mlog_setmasks);
	/* simplified high-level env var config, part 2 (LEVEL,WHICH) */
	level = getenv("IOD_DEBUG_LEVEL");
	if (level && mlog_str2pri((char *)level) == -1) {
		mlog(MLOG_WARN, "IOD_DEBUG_LEVEL error: bad level: %s", level);
		level = NULL;   /* reset to default */
	}
	ev = getenv("IOD_DEBUG_WHICH");
	if (ev == NULL) {
		if (level != NULL) {
			mlog_setmasks((char *)level, -1);/* apply to all facs */
		}
	} else {
		while (*ev) {
			start = ev;
			while (*ev != 0 && *ev != ',') {
				ev++;
			}
			snprintf(tmpbuf, sizeof(tmpbuf), "%.*s=%s",
				(int)(ev - start), start,
				(level) ? level : "DBUG");
			mlog_setmasks(tmpbuf, -1);
			if (*ev == ',') {
				ev++;
			}
		}
	}
	/* end of part 2 of simplified high-level env var config */
	mlog(IOD_INFO, "mlog init complete");
	return;
}

iod_conf_t *iod_get_conf()
{
	static pthread_mutex_t confmutex = PTHREAD_MUTEX_INITIALIZER;
	static iod_conf_t *pconf = NULL;
	char *configFileName;
	iod_conf_t *tmppconf;
	FILE *fp;

	pthread_mutex_lock(&confmutex);
	if (pconf) {
		pthread_mutex_unlock(&confmutex);
		return pconf;
	}
	/*
	 * bring up a simple mlog here so we can collect early error messages
	 * before we've got access to all the mlog config info from file.
	 * we'll replace with the proper settings once we've got the conf
	 * file loaded and the command line args parsed...
	 * XXXCDC: add code to check environment vars for non-default levels
	 */
	if (mlog_open((char *) "iodinit",
		      /* don't count the null at end of mlog_facsarray */
		      sizeof(mlog_facsarray) / sizeof(mlog_facsarray[0]) - 1,
		      MLOG_WARN, MLOG_WARN, NULL, 0, MLOG_LOGPID, 0) == 0) {
		      /* MLOG_DBG, MLOG_DBG, NULL, 0, MLOG_LOGPID, 0) == 0) { */
		setup_mlog_facnamemask(NULL);
	}

	configFileName = IOD_DEFAULT_RC;
	fp = fopen(configFileName, "r");
	if (fp == NULL) {
		mlog(MLOG_WARN, "Failed to open config file %s, "
		     "please copy iodrc to /etc",
		     configFileName);
		/* return NULL; */
		mlog(MLOG_WARN, "Use iodrc in iod top directory, "
		     "please copy iodrc to /etc next time.");
		configFileName = "../iodrc";
		fp = fopen(configFileName, "r");
		if (fp == NULL) {
			mlog(MLOG_WARN, "Failed to open config file %s.",
			     configFileName);
			return NULL;
		}
	}
	tmppconf = parse_conf(fp, configFileName, NULL);
	fclose(fp);
	if (tmppconf) {
		pconf = tmppconf;
	} else {
		pthread_mutex_unlock(&confmutex);
	}

	if (pconf) {
		setup_mlog(pconf);
	}
	pthread_mutex_unlock(&confmutex);
	return pconf;
}

/* Free some resources */
void iod_release_conf(iod_conf_t *pconf)
{
	if (pconf != NULL) {
		mlog_close();
		if (pconf->mlog_file != NULL)
			free(pconf->mlog_file);
		if (pconf->mlog_setmasks != NULL)
			free(pconf->mlog_setmasks);
		free(pconf);
	}
}
