/*
 *
 */

#include <iod_internal.h>

/**
 * Open an IOD container.
 *
 * It should be collectively called by every application process.
 * IOD will call daos_container_open to create/open DAOS container.
 *
 * \param path [IN]	container path name
 * \param hints[IN]	pointer to hints and can be NULL when no hint
 * \param mode [IN]     open mode, can be IOD_CONT_RO, IOD_CONT_WO or
 *			IOD_CONT_RW. For IOD_CONT_WO or IOD_CONT_RW can also
 *			carry a IOD_CONT_CREATE flag for creating new container.
 *			It is recommended that caller should use read-only or
 *			write-only mode if it needs not writing or reading.
 *			Always pass in IOD_CONT_RW will lose some internal
 *			optimization in IOD layer.
 * \param coh [IN/OUT]	returned container handle
 * \param event [IN]	pointer to completion event
 *
 * \return		zero on success, negative value if error
 */
iod_ret_t
iod_container_open(const char *path, iod_hint_list_t *hints, unsigned int mode,
		   iod_handle_t *coh, iod_event_t *event)
{
	IOD_ENTER;
	mlog(IOD_DAPI, "Params - path: %s, hints: %p, mode: 0x%x, event: %p",
	     path, hints, mode, event);


	IOD_EXIT;
	return 0;
}

/**
 * Close an IOD container.
 *
 * It should be collectively called by every application process.
 * All opened and un-closed IOD container will be implicitly closed inside
 * iod_finalize.
 *
 * \param coh [IN]	container handle
 * \param hints[IN]	pointer to hints and can be NULL when no hint
 * \param event [IN]	pointer to completion event
 *
 * \return		zero on success, negative value if error
 */
iod_ret_t iod_container_close(iod_handle_t coh, iod_hint_list_t *hints,
			      iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Unlink an IOD container by path.
 *
 * Any single process can call it. Caller should ensure the container is closed
 * before calling _unlink, it will fail if unlinks an un-closed container.
 *
 * \param path [IN]	container path name
 * \param force [IN]	If not zero then forcibly unlink even if container is
 *			non-empty. If set as zero then will return -ENOTEMPTY
 *			for non-empty container.
 * \param event [IN]	pointer to completion event
 *
 * \return		zero on success, negative value if error
 */
iod_ret_t iod_container_unlink(const char *path, int force, iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Get list of objects within an IOD container at a particular TID.
 *
 * It can be called by any processes, commonly should be called by single
 * process as multiple processes' concurrent calling possibly will lead to
 * performance downgrade.
 *
 * The \a tid can be gotten from iod_container_query_tids, or set as zero which
 * means to list this container's latest readable tid.
 *
 * \param coh [IN]	container handle
 * \param tid [IN]	transaction ID
 * \param filter [IN]	object type filter, only list this type object.
 *			if it is set as IOD_OBJ_ANY, then will first list ARRAY
 *			then BLOB and KV objects.
 * \param offset [IN]	offset of object ID in list, ordered by ID value from
 *			lower to higher.
 * \param num [IN]	how many objects to list
 * \param oid [IN/OUT]	returned object ID list, user provides the memory.
 * \param type [IN/OUT]	returned object type list, user provides the memory.
 * \param name [IN/OUT]	object name, can be empty for anonymous object.
 *			Caller should provide the buffer list for name (with
 *			length as IOD_OBJ_NAME_MAXLEN), the first name's buffer
 *			address is (name), the second is (name +
 *			IOD_OBJ_NAME_MAXLEN) and so on.
 *			Can be NULL if caller does not care the name.
 * \param event [IN]	pointer to completion event
 * \return		zero on success, negative value if error
 */
iod_ret_t iod_container_list_obj(iod_handle_t coh, iod_trans_id_t tid,
				 iod_obj_type_t filter, iod_off_t offset,
				 iod_size_t num, iod_obj_id_t *oid,
				 iod_obj_type_t *type, char *name,
				 iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Create snapshot for a IOD container based on its latest readable status.
 *
 * Only single rank can call this routine to create a container snapshot.
 * IOD will migrate latest readable TID to DAOS and create DAOS container
 * snapshot by calling daos_container_snapshot.
 *
 * \param coh [IN]		container handle
 * \param snapshot [IN]		name of snapshot
 * \param hints[IN]		pointer to hints and can be NULL when no hint
 * \param event [IN]		pointer to completion event
 *
 * \return			zero on success, negative value if error
 */
iod_ret_t iod_container_snapshot(iod_handle_t coh, const char *snapshot,
				 iod_hint_list_t *hints, iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}
