/*
 *
 */

#include <iod_internal.h>

/**
 * Create one IOD object.
 *
 * Only single process can call it. Common usage is one rank creates object and
 * shares the returned object ID to other ranks.
 *
 * Permissions inherited from parent container, when creates array object it
 * will set the array's default dimension sequence(same as logical dimension).
 *
 * \param coh [IN]		container handle
 * \param tid [IN]		transaction ID
 * \param hints[IN]		pointer to hints and can be NULL when no hint
 * \param type[IN]		object type
 * \param name[IN]		optional object name, will be ignored for KV
 *				object as KV object is nameless.
 * \param array_struct[IN]	array structure, only meaningful for array.
 * \param oid[OUT]		returned object ID
 * \param event [IN]		pointer to completion event
 *
 * \return			zero on success, negative value if error
 */
iod_ret_t iod_obj_create(iod_handle_t coh, iod_trans_id_t tid,
			 iod_hint_list_t *hints, iod_obj_type_t type,
			 const char *name, iod_array_struct_t *array_struct,
			 iod_obj_id_t *oid, iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Create a set of IOD objects within one transaction.
 * It is a wrapper inside which looply calls iod_obj_create.
 *
 * \param coh [IN]		container handle
 * \param tid [IN]		transaction ID
 * \param num [IN]		how many objects to create
 * \param obj_create [IN/OUT]	object create structure list pointer
 * \param event [IN]		pointer to completion event
 *
 * \return			zero on success, negative value if error
 */
iod_ret_t iod_obj_create_list(iod_handle_t coh, iod_trans_id_t tid,
			      iod_size_t num, iod_obj_create_t *obj_create,
			      iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Open one IOD object for writing.
 *
 * Every rank which wants to do I/O for the object should call it to get a
 * opened object handle.
 *
 * The iod_obj_create will set the object's default layout, after open
 * the caller can change it by calling iod_obj_set_layout.
 *
 * \param coh [IN]	container handle
 * \param oid [IN]	object ID
 * \param hints[IN]	pointer to hints and can be NULL when no hint
 * \param oh [IN/OUT]	object handle pointer
 * \param event [IN]	pointer to completion event
 *
 * \return		zero on success, negative value if error
 */
iod_ret_t iod_obj_open_write(iod_handle_t coh, iod_obj_id_t oid,
			     iod_hint_list_t *hints, iod_handle_t *oh,
			     iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Open a set of IOD objects for writing.
 * It is a wrapper inside which looply calls iod_obj_open_write.
 *
 * \param coh [IN]		container handle
 * \param num [IN]		number of objects to open
 * \param open_write [IN/OUT]	pointer list of packed parameters
 * \param event [IN]		pointer to completion event
 *
 * \return			zero on success, negative value if error
 */
iod_ret_t iod_obj_open_write_list(iod_handle_t coh, iod_size_t num,
				  iod_obj_open_t *open_write,
				  iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Open one IOD object for read.
 *
 * Every rank which wants to do I/O for the object should call it to get a
 * opened object handle.
 *
 * \param coh [IN]	container handle
 * \param oid [IN]	object ID
 * \param hints[IN]	pointer to hints and can be NULL when no hint
 * \param oh [OUT]	object handle pointer
 * \param event [IN]	pointer to completion event
 *
 * \return		zero on success, negative value if error
 */
iod_ret_t iod_obj_open_read(iod_handle_t coh, iod_obj_id_t oid,
			    iod_hint_list_t *hints, iod_handle_t *oh,
			    iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Open a set of IOD objects for read.
 * It is a wrapper inside which looply calls iod_obj_open_read.
 *
 * \param coh [IN]		container handle
 * \param num [IN]		how many object to read
 * \param open_read [IN/OUT]	pointer list of packed parameters
 * \param event [IN]		pointer to completion event
 *
 * \return			zero on success, negative value if error
 */
iod_ret_t iod_obj_open_read_list(iod_handle_t coh, iod_size_t num,
				 iod_obj_open_t *open_read, iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Close one IOD object.
 *
 * Every rank which has opened object should close it at appropriate time.
 * User should call close in time when this rank does not intend to do I/O for
 * that object, this can facilitate IOD's internal optimization and resource
 * freeing.
 *
 * \param oh [OUT]	object handle
 * \param hints[IN]	pointer to hints and can be NULL when no hint
 * \param event [IN]	pointer to completion event
 *
 * \return		zero on success, negative value if error
 */
iod_ret_t iod_obj_close(iod_handle_t oh, iod_hint_list_t *hints,
			iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Close a set of IOD objects within one container.
 * It is a wrapper inside which looply calls iod_obj_close.
 *
 * \param coh [IN]		container handle
 * \param num [IN]		how many objects to close
 * \param obj_close [IN/OUT]	pointer list of packed parameters
 * \param event [IN]		pointer to completion event
 *
 * \return			zero on success, negative value if error
 */
iod_ret_t iod_obj_close_list(iod_handle_t coh, iod_size_t num,
			     iod_obj_close_t *obj_close, iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Set one IOD object's layout on storage system.
 *
 * This routine is only for setting object's layout on BB or central storage.
 * It will not start/trigger data movement. Only single rank should call it.
 *
 * Common usage is user can first set the layout, when iod_trans_persist IOD
 * will based on this setted layout to determine data placement when migrating
 * this object to DAOS. At that time IOD will migrate object's data with the
 * dimension sequence of \a layout.dims_seq, all split shards will be
 * round-robin placed on \a layout.target_num's targets with stripe size of
 * \a layout.stripe_size.
 *
 * \param oh [IN]		object handle
 * \param tid [IN]		transaction ID
 * \param hints[IN]		pointer to hints and can be NULL when no hint
 * \param layout [IN]		passed in layout of the object
 * \param event [IN]		completion event pointer list
 *
 * \return			zero on success, negative value if error
 */
iod_ret_t iod_obj_set_layout(iod_handle_t oh, iod_trans_id_t tid,
			     iod_hint_list_t *hints, iod_layout_t *layout,
			     iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Query one IOD object's layout on storage system.
 *
 * Any rank can call it, commonly single rank calls it to query.
 *
 * \param oh [IN]		object handle
 * \param tid [IN]		transaction ID, must be readable
 * \param layout [IN/OUT]	returned object layout
 * \param event [IN]		completion event pointer list
 *
 * \return			zero on success, negative value if error
 */
iod_ret_t iod_obj_get_layout(iod_handle_t oh, iod_trans_id_t tid,
			     iod_layout_t *layout, iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Unlink one IOD object by ID.
 *
 * Caller should close it before unlink, it will fail if that object is in open.
 * Only single rank should call it.
 *
 * \param coh [IN]	container handle
 * \param oid [IN]	object ID
 * \param tid [IN]	transaction ID, object will be un-existed at TIDs equal
 *			or greater than this \a tid, but will keep existed at
 *			TIDs lower than this \a tid.
 * \param event [IN]	pointer to completion event
 *
 * \return		zero on success, negative value if error
 */
iod_ret_t iod_obj_unlink(iod_handle_t coh, iod_obj_id_t oid, iod_trans_id_t tid,
			 iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Unlink a set of IOD objects by IDs, all in one transaction.
 *
 * \param coh [IN]		container handle
 * \param tid [IN]		transaction ID
 * \param num [IN]		how many objects to unlink
 * \param obj_unlink [IN/OUT]	pointer list of packed parameters
 * \param event [IN]		pointer to completion event
 *
 * \return		zero on success, negative value if error
 */
iod_ret_t iod_obj_unlink_list(iod_handle_t coh, iod_trans_id_t tid,
			      iod_size_t num, iod_obj_unlink_t *obj_unlink,
			      iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Set the 32 bytes scratchpad of an object.
 * IOD supports one 32 bytes length scratchpad for every object. User can use
 * it to store object's child-object ID or something else. For IOD, it is just
 * 32 bytes stream.
 *
 * It will fail if the object handle has no writing permission.
 * Any rank can call it, commonly single rank calls it to set.
 *
 * \param oh [IN]		object handle
 * \param tid [IN]		transaction ID
 * \param scratch[IN]		passed in 32bytes stream of scratchpad
 * \param cs[IN]		passed in checksum pointer for the scratch set
 * \param event [IN]		pointer to completion event
 *
 * \return			zero on success, negative value if error
 */
iod_ret_t iod_obj_set_scratch(iod_handle_t oh, iod_trans_id_t tid,
			      const char* scratch, iod_checksum_t *cs,
			      iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Set the 32 bytes scratchpad for a set of objects.
 * It is a wrapper inside which looply calls iod_obj_set_scratch.
 *
 * \param coh [IN]		container handle
 * \param tid [IN]		transaction ID
 * \param num [IN]		how many objects to unlink
 * \param scratch_set[IN/OUT]	pointer to iod_obj_scratch_t list
 * \param event [IN]		pointer to completion event
 *
 * \return			zero on success, negative value if error
 */
iod_ret_t iod_obj_set_scratch_list(iod_handle_t coh, iod_trans_id_t tid,
				   iod_size_t num,
				   iod_obj_scratch_t *scratch_set,
				   iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Get the 32 bytes scratchpad of an object.
 * IOD supports one 32 bytes length scratchpad for every object. User can use
 * it to store object's child-object ID or something else. For IOD, it is just
 * 32 bytes stream.
 *
 * Any rank can call it, commonly single rank calls it to get.
 *
 * \param oh [IN]		object handle
 * \param tid [IN]		transaction ID
 * \param scratch[IN/OUT]	32bytes stream of scratchpad
 * \param cs[IN/OUT]		returned checksum for the scratchpad
 * \param event [IN]		pointer to completion event
 *
 * \return			zero on success, negative value if error
 */
iod_ret_t iod_obj_get_scratch(iod_handle_t oh, iod_trans_id_t tid,
			      const char* scratch, iod_checksum_t *cs,
			      iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Get the 32 bytes scratchpad for a set of objects.
 * It is a wrapper inside which looply calls iod_obj_get_scratch.
 *
 *
 * \param coh [IN]		container handle
 * \param tid [IN]		transaction ID
 * \param num [IN]		how many objects to unlink
 * \param scratch_get[IN/OUT]	pointer to iod_obj_scratch_t list
 * \param event [IN]		pointer to completion event
 *
 * \return			zero on success, negative value if error
 */
iod_ret_t iod_obj_get_scratch_list(iod_handle_t coh, iod_trans_id_t tid,
				   iod_size_t num,
				   iod_obj_scratch_t *scratch_get,
				   iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Purge a object for an IOD container from BB.
 *
 * Only single rank can call this routine to purge one object from BB.
 * This is commonly used after migrating done and that object of \a tid is
 * un-needed to be kept on BB. IOD won't do auto-purging, user should explicitly
 * call the purge at appropriate time to free BB's storage space. Purging is at
 * object level, IOD will purge all lower TIDs and this TID \a tid.
 *
 * User can only call purge with the \a tid that ever passed in by
 * iod_trans_persist. If it is called before the completion of iod_trans_persist
 * IOD will delay the real purging after the success completion of
 * iod_trans_persist.
 *
 * \param oh [IN]	object handle
 * \param tid [IN]	transaction ID
 * \param hints[IN]	pointer to hints and can be NULL when no hint
 * \param event [IN]	pointer to completion event
 *
 * \return		zero on success, negative value if error
 */
iod_ret_t iod_obj_purge(iod_handle_t oh, iod_trans_id_t tid,
			iod_hint_list_t *hints, iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Fetch/pre-stage one object from central storage to BB.
 *
 * Only single rank can call this routine to fetch one object from central
 * storage to BB.
 *
 * \param oh [IN]		object handle
 * \param tid [IN]		transaction ID
 * \param hints[IN]		pointer to hints and can be NULL when no hint
 * \param slab [IN]		the hyperslab within this object
 *				will be ignored for KV object.
 * \param layout [IN]		passed in layout placement on BB,
 *				can be NULL if does not change layout on BB.
 * \param new_tid[IN/OUT]	returned new TID, will be same as passed in
 *				\a tid if does not change the \a layout on BB.
 * \param event [IN]		completion event pointer list
 *
 * \return			zero on success, negative value if error
 */
iod_ret_t iod_obj_fetch(iod_handle_t oh, iod_trans_id_t tid,
			iod_hint_list_t *hints, iod_hyperslab_t *slab,
			iod_layout_t *layout, iod_trans_id_t *new_tid,
			iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Create a replica for one object from BB to BB.
 *
 * Only single rank can call this routine to create a replica for existed object
 * on BB. Later user should explicitly call iod_obj_purge to purge this replica
 * when it is not needed.
 *
 * \param oh [IN]		object handle
 * \param tid [IN]		transaction ID
 * \param hints[IN]		pointer to hints and can be NULL when no hint
 * \param this_only [IN]	un-zero means only create a replica for this TID
 *				incremented data; Zero means replicate all TIDs'
 *				data till this \a tid.
 * \param layout [IN]		passed in layout placement on BB, if it is same
 *				as previous layout IOD will not do anything.
 * \param new_tid[IN/OUT]	returned new TID, will be same as passed in
 *				\a tid if does not change the \a layout on BB.
 * \param event [IN]		completion event pointer list
 *
 * \return			zero on success, negative value if error
 */
iod_ret_t iod_obj_replica(iod_handle_t oh, iod_trans_id_t tid,
			  iod_hint_list_t *hints, unsigned int this_only,
			  iod_layout_t *layout, iod_trans_id_t *new_tid,
			  iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}
