/*
 *
 */

#include <iod_internal.h>

/**
 * Query one IOD container's TIDs status.
 *
 * Any rank can call it, commonly single rank calls it to query and can share
 * it to other ranks.
 *
 * Caller should be aware that there could be race condition here - when this
 * call returns, the underneath TID status possibly being changed as other work
 * is ongoing.
 *
 * \param coh [IN]		container handle
 * \param tids [IN/OUT]		pointer to container tid status
 * \param event [IN]		pointer to completion event
 *
 * \return			zero on success, negative value if error
 */
iod_ret_t iod_container_query_tids(iod_handle_t coh, iod_container_tids_t *tids,
				   iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Query one TID's status.
 *
 * Caller should be aware that there could be race condition here - when this
 * call returns, the underneath TID status possibly being changed as other work
 * is ongoing.
 *
 * \param coh [IN]		container handle
 * \param tid [IN]		TID to query
 * \param status [IN/OUT]	pointer to transaction status
 * \param event [IN]		pointer to completion event
 *
 * \return			zero on success, negative value if error
 */
iod_ret_t iod_trans_query(iod_handle_t coh, iod_trans_id_t tid,
			  iod_trans_status_t *status, iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Start a new transaction for an IOD container.
 *
 * User should start a transaction for either writing(\a mode == IOD_TRANS_WR)
 * or reading(\a mode == IOD_TRANS_RD). IOD does not allow to read and write a
 * TID at the same time. iod_container_query_tids can help to get an appropriate
 * TID for writing or reading, user also can pass in IOD_TID_UNKNOWN as
 * explained below.
 *
 * For writing, user should pass in an appropriate TID, for the same TID all
 * callers must call it with same \a num_ranks. User can get the appropriate
 * TID by either:
 * 1) calling iod_container_query_tids, the TID higher than latest_wrting is
 *    writable - common use case is to start (latest_wrting+1); or
 * 2) passing in IOD_TID_UNKNOWN, IOD will set the \a tid as (latest_wrting+1)
 *    and allow user to write to it.
 * If two independent groups need to start new transaction for writing, then
 * commonly they should pass in IOD_TID_UNKNOWN to avoid TID confliction.
 *
 * For reading, user should pass in an appropriate TID, for the same TID all
 * callers must call it with same \a num_ranks. User can get the appropriate
 * TID by either:
 * 1) calling iod_container_query_tids, IOD can ensure the read consistency of
 *    TIDs between lowest_durable and latest_rdable (note that there might be
 *    some aborted/invalid TIDs which are not readable); or
 * 2) passing in IOD_TID_UNKNOWN, IOD will set the \a tid as latest_rdable and
 *    allow user to read from it.
 *    Hints: key "lowest_readable" value "true" means start by lowest readable
 *           TID. For example, on BB there are readable TIDs 8, 9, 10, 11. This
 *           hint will cause start by TID 8.
 *	     If no hints or other invalid hints then IOD will slip to latest
 *	     readable TID. In the above example will start by TID 11.
 *
 * A ref-count will be taken when this call finishes sucessfully, and will be
 * released after iod_trans_finish or iod_trans_slip's sucess completion.
 *
 * \a num_ranks is the number of participators for this transaction. IOD
 * supports two kinds of transaction status synchronization and consistency
 * ensuring mechanism:
 * 1) application does the transaction status synchronization and consistency
 *    ensuring. CN ranks will need to select one transaction leader. The leader
 *    starts and finishes/slips the TID, other ranks can participate in the TID
 *    after leader having started it, and the leader should ensure all other
 *    ranks have finished I/O operations within this TID before it finish/slip
 *    this TID. This is the method similar as DAOS epoch's requirement.
 *    User should pass in \a num_ranks as zero for this method.
 * 2) IOD internally does the transaction status synchronization and consistency
 *    ensuring. CN ranks only need to start and finish/slip this TID seperately
 *    and independently. IOD will do lots of internal P2P message passing for
 *    status synchronization. As IOD does not know CN ranks' process group
 *    information, so IOD cannot use group collective communication. When
 *    \a num_ranks is large, the status synchronizations will introduce
 *    considerable overhead and latency at IOD layer.
 *    User should pass in \a num_ranks as the number of CN ranks that will
 *    participate in this transaction for this method. IOD_TID_UNKNOWN cannot be
 *    used for this method.
 * Commonly application should use method 1) to participate transaction.
 *
 * \param coh [IN]		container handle
 * \param tid [IN/OUT]		pointer to transaction ID
 * \param hints[IN]		pointer to hints and can be NULL when no hint
 * \param num_ranks[IN]		count of participators(CN ranks)
 * \param mode[IN]		either IOD_TRANS_RD or IOD_TRANS_WR,
 *				will return -EINVAL if passes in other value.
 * \param event [IN]		pointer to completion event
 *
 * \return			zero on success, negative value if error
 */
iod_ret_t iod_trans_start(iod_handle_t coh, iod_trans_id_t *tid,
			  iod_hint_list_t *hints, unsigned int num_ranks,
			  unsigned int mode, iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Slip a transaction for an IOD container.
 *
 * The slip is like a combination of "iod_trans_finish(old_tid) +
 * iod_trans_start(new_tid)". User should pass in an determined and appropriate
 * \a tid to slip(cannot be IOD_TID_UNKNOWN). Slip can only be done when the
 * num_ranks is same between old_tid and new_tid, or user should call
 * iod_trans_finish(old_tid) and iod_trans_start(new_tid) seperately with the
 * different num_ranks.
 *
 * If \a tid is for writing, IOD will do "iod_trans_finish(*tid) +
 * iod_trans_start(next_writable_tid)". After completion, *tid is set as next
 * writable tid which commonly is (*tid + 1).
 *
 * If \a tid is for reading, IOD will do "iod_trans_finish(*tid) +
 * iod_trans_start(next_readable_tid)". After completion, *tid is set as next
 * readable tid.
 * Hints: key "adjacent_readable" value "true" means slip to adjacent readable
 *        TID. For example, current (*tid) is 8, and have higher readable TIDs
 *        9, 10, 11. This hints will cause slip to TID 9.
 *	  If no hints or other invalid hints then IOD will slip to latest
 *	  readable TID. In the above example will slip to TID 11.
 *	  This hints is only meaningful for reading.
 *
 * \param coh [IN]		container handle
 * \param tid [IN/OUT]		pointer to transaction ID
 * \param hints[IN]		pointer to hints and can be NULL when no hint
 * \param event [IN]		pointer to completion event
 *
 * \return			zero on success, negative value if error
 */
iod_ret_t iod_trans_slip(iod_handle_t coh, iod_trans_id_t *tid,
			 iod_hint_list_t *hints, iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}

/**
 * Finish a transaction for an IOD container.
 *
 * For writing, if one tid's all participators have completed the
 * iod_trans_finish, and this tid's all former started tids have become
 * readable or aborted, then this tid become readable in burst buffer. The
 * success completion of iod_trans_finish means this \a tid become readable
 * on BB. However, to ensure one tid's data to be durable on central storage,
 * user should call iod_trans_persist to migrate/protect it.
 *
 * The \a abort is to abort an transaction, all writings/updatings in this
 * transaction will be rolled back after success aborting.
 *
 * For reading, this routine is only for releasing the ref-count taken by
 * iod_trans_start(or _slip).
 *
 * \param coh [IN]	container handle
 * \param tid [IN]	transaction ID
 * \param abort [IN]	abort the transaction, only meaningful for writing.
 *			1) Zero for commonly finish;
 *			2) IOD_TRANS_ABORT_ALL for aborting this TID and all
 *			   higher TIDs in writing;
 *			3) IOD_TRANS_ABORT_SINGLE for aborting only this TID.
 *			Other value will return -EINVAL.
 * \param hints[IN]	pointer to hints and can be NULL when no hint
 * \param event [IN]	pointer to completion event
 *
 * \return		zero on success, negative value if error
 */
iod_ret_t iod_trans_finish(iod_handle_t coh, iod_trans_id_t tid,
			   iod_hint_list_t *hints, int abort,
			   iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}


/**
 * Persist/protect a transaction for an IOD container.
 *
 * Only single rank should call this routine to make this transaction be
 * migrated to central storage. User can only persist a readable TID.
 *
 * IOD will make all lower readable TIDs and this TID's state be persist. If
 * user want to set/change some objects' layout on central storage, it should
 * call iod_obj_set_layout before calling iod_trans_persist. After persist's
 * completion this \a tid become durable on central storage. The previous
 * readable TID is still on BB, IOD will not automatically purge it.
 *
 * \param coh [IN]	container handle
 * \param tid [IN]	transaction ID
 * \param hints[IN]	pointer to hints and can be NULL when no hint
 * \param event [IN]	pointer to completion event
 *
 * \return		zero on success, negative value if error
 */
iod_ret_t iod_trans_persist(iod_handle_t coh, iod_trans_id_t tid,
			    iod_hint_list_t *hints, iod_event_t *event)
{
	IOD_ENTER;


	IOD_EXIT;
	return 0;
}
